/*	$Id: libimages3d.h 2007/11/23 19:24:31 mallet $
 *
 * Copyright (c) 2006-2007 LAAS/CNRS                  --  Thu Aug  3 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef CALIFE_LIBIMAGES3D_H
#define CALIFE_LIBIMAGES3D_H

#include <stdio.h>

#include "t3d/t3d.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Les structures */
/* Diffrentes numrations pour les flags des points (aussi utilises pour
   l'entte des images). Les noms des flags ainsi que leurs valeurs sont
   censes tre explicites. */
typedef enum {BAD_PT3D,
	      GOOD_PT3D} state_info;
typedef enum {NIL_NORMALES,
	      NORMALES_ALLOUEES,
	      NORMALES_CALCULEES} normales_info;
typedef enum {NIL_TYPE_NORMALES,
	      QUATRE_VOISINS,
	      HUIT_VOISINS} type_normales;
typedef enum {NIL_COVARIANCES,
	      COVARIANCES_ALLOUEES,
	      COVARIANCES_CALCULEES} covariances_info;
typedef enum {NIL_INFOS_LOCALES,
	      INFOS_LOCALES_ALLOUEES,
	      INFOS_LOCALES_CALCULEES} local_info;
typedef enum {NIL_COORDONNEES,
	      COORDONNEES_CARTESIENNES,
	      COORDONNEES_CYLINDRIQUES,
	      COORDONNEES_SPHERIQUES} type_coordonnees;

/* Diffrentes numrations pour l'entte de l'image (pas utilises pour les
   points), qui donnent des infos sur son tat, son origine ou n'importe quoi
   d'autre */
typedef enum {NIL_BORNES,
	      BORNES_CALCULEES} bornes_info;
typedef enum {NIL_ROBOT,
	      ROBOT_ADAM,
	      ROBOT_HILARE,
	      ROBOT_HILARE2,
	      ROBOT_LAMA,
	      ROBOT_EVE,
	      ROBOT_INCONNU} type_robot;
typedef enum {NIL_CAPTEUR,
	      CAPTEUR_LASER1,
	      CAPTEUR_LASER2,
	      CAPTEUR_STEREO,
	      CAPTEUR_STEREO_GEROMS94,
	      CAPTEUR_LASER_SIMULATION,
	      CAPTEUR_STEREO_SIMULATION} type_capteur;
typedef enum {NIL_REPERE,
	      REPERE_CAPTEUR,
	      REPERE_AUTRE} type_repere;
typedef enum {NIL_POINTS,
	      POINTS_ALLOUES,
	      POINTS_CALCULES} points_info;
typedef enum {NIL_REDUCTION,
	      REDUCTION} reduction_info;
typedef enum {NIL_SOUS_IMAGE,
	      SOUS_IMAGE} sous_image_info;

/* typedef enum {NIL_SOUS_IMAGE, SOUS_IMAGE} sous_image_info; */


/* Les tailles des tableaux des covariances et du vecteur normal */
#define IM3D_COVARIANCES_SIZE 6
#define IM3D_NORMALES_SIZE 3

/* Enttes des fichiers de sauvegarde de la structure 'data_im3d' (FF = file
 * format) */
#define FF_DATA_IM3D_ASCII	"DATA_IM3D_ascii_file_format"
#define FF_DATA_IM3D_BINARY	"DATA_IM3D_binary_file_format"

/***********************************************************************
 *
 * La structure 'data_pt3d'.
 *
 ***********************************************************************/

typedef struct data_pt3d
{
  /* Un champ de bit dans lequel on peut mettre tout un tas
     d'informations. Voir dans la squence des numrations ci-dessus la
     signification des valeurs correspondantes */
  unsigned       state:1;	/* L'tat du point */
  unsigned       repere:2;	/* Le repre des ses coordonnes */
  unsigned       coordonnees:2;	/* Systme de coordonnes */
  unsigned       normales:2;	/* Normale */
  unsigned       nb_voisins_normales:2; /* Type de calcul de normales */
  unsigned       covariances:2;	/* Covariances */
  unsigned       local:2;	/* Flags sur les infos locales */
  unsigned       flag_perso:3;	/* Gros flag (valeurs de 0  8) dans lequel
				   chacun met ce qu'il veut */

  float          coord_1;
  float          coord_2;
  float          coord_3;	/* Les trois coordonnes, dans cet ordre :
				   (x,y,z) en cartsiennes, (rho, theta, phi)
				   en sphriques et (r, theta, z) en
				   cylindriques */

  float*         covariances_tab; /* Pointeur vers la matrice de covariance (6
				     floats) */
  float*         normales_tab;	/* Pointeur vers le vecteur normal (3 floats) */

  void*          local_tab;	/* Pointeur vers d'autres infos */

}  DATA_PT3D;


/***********************************************************************
 *
 * L'entte de la structure image de 'point3d'. Il n'y a aucun pointeur dans
 * cette structure, ce qui vite des fcts de cration/destruction spcifiques
 (ces oprations sont faites lors de la cration/destruction de la structure
 * DATA_IM3D)
 *
 ***********************************************************************/

typedef struct data_im3d_header {
  /* On commence par tout un tas de caractristiques gnrales de l'image */
  type_robot        robot;	/* Le robot avec lequel a t acquise l'image */
  type_capteur      capteur;	/* Le capteur avec lequel a t acquise l'image */
  type_repere       repere;	/* Indique si les pts sont exprims dans le
				   repre capteur ou autre */
  type_coordonnees  coordonnees; /* Le systeme de coordonnes dans lequel sont
				    exprims les points de l'image */
  reduction_info    reduction;	/* Indique si l'image  t
				   sous-chantillonne dans la cours de sa vie
				*/
  sous_image_info   sous_image;	/* Indique si cette image a t extraite d'une
				   image plus grande */

  /* Suivent des infos sur ce qu'on a fait subir aux points de l'image. Voir
     ci-dessus pour le sens des champs */
  points_info       points;
  normales_info     normales;
  type_normales     nb_voisins_normales;
  covariances_info  covariances;
  local_info        local;
  bornes_info       bornes;

  /* Infos sur la taille de l'image. Toutes ces complications sont l pour que
     l'on puisse reduire et sous-chantillonner une image dans n'importe quel
     ordre, autant de fois qu'on veut, et que la matrice de calibration ait
     toujours un sens */
  int               nblig_initial;
  int               nbcol_initial; /* Nombre de lignes et de colonnes au
				      original : utile pour que la matrice de
				      calibration ait toujours un sens */
  int               nblig;
  int               nbcol;	/* Nombre de lignes et de colonnes effectifs
				   dans l'image (ie. aprs ventuels
				   sous-chantillonages et rductions) */
  int               i_min;
  int               i_max;
  int               j_min;
  int               j_max;	/* Coordonnes de la sous-image PAR RAPPORT 
				   LA TAILLE INITIALE: donc si l'image a t
				   rduite, on n'a pas nblig = i_max - i_min,
				   mais nblig = (i_max - i_min) /
				   taux_reduction. En consquence, il ne faut
				   jamais boucler entre les min et max mais
				   entre 0 et nblig[col]. */
  int               taux_reduction; /* Indique le taux de rduction de l'image
				       (si elle a t sous-chantillonne) */


  /* Infos concernant la gomtrie du capteur et les repres dans lequels
     sont exprims les points */
  double            mat_intrinseque[16];
				/* Matrice intrinsque (lorsque le capteur est
				   la stro) */
  T3D		    tr_capteur_2_repere; /* La tr3d qui donne la
					    transformation capteur->repere
					    courant  */




  /* Autres infos */
  int               size_local;	/* Taille de ce qu'on a ventuellement allou
				   dans le champ 'local_tab' des points */

  float             max_coord_1;
  float             min_coord_1;
  float             max_coord_2;
  float             min_coord_2;
  float             max_coord_3;
  float             min_coord_3; /* Bornes sur les 3 coordonnes */

} DATA_IM3D_HEADER;


/***********************************************************************
 *
 * La structure image de 'point3d' : un entte + un tableau de points, et
 * ventuellement des tableaux d'autres infos (normales, co-variances et
 * autres)
 *
 ***********************************************************************/

typedef struct data_im3d
{
  /* L'entte */
  DATA_IM3D_HEADER header;

  /* Les donnes */
  DATA_PT3D*    points_tab;	/* le tableau des points3d... */
  float*        covariances_tab; /* Les covariances associes */
  float*        normales_tab;	/* Les normales associes */
  void*         local_tab;	/* Le tableaux des ventuels champs locaux */
}  DATA_IM3D;

/* macro d'acces  un point de l'im3d */
#define data_im3d_pt(a,i,j) (((a)->points_tab) + ((i) * (((a)->header).nbcol)) + (j))


/* Les erreurs */
#define im3derr(m,n)	((((m) & 0x9FFF) << 16) | ((n) & 0xFFFF))

#define images3dId	510

#define ERR_IM3D_OK					0

/* Les erreurs sur les paramtres passs aux fonctions */
#define ERR_IM3D_INDICES_CALCUL_NORMALE_POINT		im3derr(images3dId, 00)
#define ERR_IM3D_TYPE_CALCUL_NORMAL			im3derr(images3dId, 01)
#define ERR_IM3D_TAILLE					im3derr(images3dId, 02)
#define ERR_IM3D_INDICES				im3derr(images3dId, 03)
#define ERR_IM3D_TAILLE_SOUS_IMAGE			im3derr(images3dId, 04)
#define ERR_IM3D_TAUX_REDUCTION				im3derr(images3dId, 05)
#define ERR_IM3D_TAILLES_IMAGES				im3derr(images3dId, 06)

/* Le code suivant sert pour un peu tout */
#define ERR_IM3D_PARAMETRE				im3derr(images3dId, 9)

/* Les erreurs sur le contenu des images passes aux fonctions */
#define ERR_IM3D_TYPE_COORDONNEES			im3derr(images3dId, 10)
#define ERR_IM3D_POINTS_NON_ALLOUES			im3derr(images3dId, 11)
#define ERR_IM3D_POINTS_NON_CALCULES			im3derr(images3dId, 12)
#define ERR_IM3D_NORMALES_NON_CALCULEES			im3derr(images3dId, 13)
#define ERR_IM3D_COVARIANCES_NON_CALCULEES		im3derr(images3dId, 14)

/* Les erreurs sur les allocations : impossibilit d'allouer */
#define ERR_IM3D_MALLOC_PT3D				im3derr(images3dId, 20)
#define ERR_IM3D_MALLOC_PT3D_COVARIANCES		im3derr(images3dId, 21)
#define ERR_IM3D_MALLOC_PT3D_NORMALES			im3derr(images3dId, 22)
#define ERR_IM3D_MALLOC_PT3D_INFOS_LOCALES		im3derr(images3dId, 23)

#define ERR_IM3D_MALLOC					im3derr(images3dId, 25)
#define ERR_IM3D_MALLOC_TR3D				im3derr(images3dId, 26)
#define ERR_IM3D_MALLOC_POINTS				im3derr(images3dId, 27)
#define ERR_IM3D_MALLOC_COVARIANCES			im3derr(images3dId, 28)
#define ERR_IM3D_MALLOC_NORMALES			im3derr(images3dId, 29)
#define ERR_IM3D_MALLOC_INFOS_LOCALES			im3derr(images3dId, 30)

#define ERR_IM3D_CRR8					im3derr(images3dId, 31)


/* Les erreurs sur les allocations : allocations dj faites */
#define ERR_IM3D_POINTS_DEJA_ALLOUES			im3derr(images3dId, 40)
#define ERR_IM3D_NORMALES_DEJA_ALLOUEES			im3derr(images3dId, 41)
#define ERR_IM3D_COVARIANCES_DEJA_ALLOUEES		im3derr(images3dId, 42)
#define ERR_IM3D_INFOS_LOCALES_DEJA_ALLOUEES		im3derr(images3dId, 43)

#define ERR_IM3D_NORMALES_PT3D_DEJA_ALLOUEES		im3derr(images3dId, 45)
#define ERR_IM3D_COVARIANCES_PT3D_DEJA_ALLOUEES		im3derr(images3dId, 46)
#define ERR_IM3D_INFOS_LOCALES_PT3D_DEJA_ALLOUEES	im3derr(images3dId, 47)

/* Les erreurs d'ouverture et de format de fichiers */
#define ERR_IM3D_OPEN_GEROMS94_FILE			im3derr(images3dId, 60)
#define ERR_IM3D_FORMAT_GEROMS94_FILE			im3derr(images3dId, 61)
#define ERR_IM3D_READ_GEROMS94_FILE			im3derr(images3dId, 62)

#define ERR_IM3D_FORMAT_ASCII_FILE			im3derr(images3dId, 65)
#define ERR_IM3D_OPEN_ASCII_FILE			im3derr(images3dId, 66)
#define ERR_IM3D_ASCII_FILE_EXISTE			im3derr(images3dId, 67)
#define ERR_IM3D_READ_HEADER				im3derr(images3dId, 68)
#define ERR_IM3D_READ_ASCII_FILE			im3derr(images3dId, 69)
#define ERR_IM3D_READ_BINARY_FILE			im3derr(images3dId, 70)


/* Un cas de sortie qui ne devrait jamais arriver normalement */
#define ERR_IM3D_CAS_IMPOSSIBLE				im3derr(images3dId, 98)

/* Ben oui */
#define ERR_IM3D_FCT_NON_IMPLEMENTEE			im3derr(images3dId, 99)


int alloue_im3d_covariances ( DATA_IM3D* im3d );
int alloue_im3d_infos_locales ( DATA_IM3D* im3d, int size_infos );
int alloue_im3d_normales ( DATA_IM3D* im3d );
int alloue_im3d_points ( DATA_IM3D* im3d, int nblig, int nbcol );
int alloue_pt3d_covariances ( DATA_PT3D* pt3d );
int alloue_pt3d_infos_locales ( DATA_PT3D* pt3d, int size_infos );
int alloue_pt3d_normales ( DATA_PT3D* pt3d );

int calcule_normale_data_pt3d ( DATA_IM3D* im3d, int i, int j, type_normales type_calcul );
int change_repere_data_im3d ( DATA_IM3D* im3d, T3D* transfo );
int change_repere_data_pt3d ( DATA_PT3D* pt3d, T3D* transfo );

DATA_IM3D* copy_data_im3d ( DATA_IM3D* im3d );
int copy_data_pt3d ( DATA_PT3D* point, DATA_PT3D* copie, int size_local );
int copy_im3d ( DATA_IM3D *im3d, DATA_IM3D *copie );
int cree_data_im3d ( DATA_IM3D **im3d, int nblig, int nbcol, int covar, int normales, int size_local );
int cree_data_pt3d ( DATA_PT3D** pt3d, int covar, int normales, int size_local );

int data_im3d_cart_to_cyl ( DATA_IM3D* im3d );
int data_im3d_cart_to_spher ( DATA_IM3D* im3d );

int data_im3d_cyl_to_cart ( DATA_IM3D* im3d );
int data_im3d_cyl_to_spher ( DATA_IM3D* im3d );
int data_im3d_minmax ( DATA_IM3D* im3d );
int data_im3d_normales ( DATA_IM3D* im3d, type_normales type_calcul );

int data_im3d_spher_to_cart ( DATA_IM3D* im3d );
int data_im3d_spher_to_cyl ( DATA_IM3D* im3d );

void del_data_im3d ( DATA_IM3D** im3d );
void del_data_pt3d ( DATA_PT3D** pt3d );

int dup_data_pt3d ( DATA_PT3D* point, DATA_PT3D** copie, int size_local );
int dup_im3d ( DATA_IM3D *im3d, DATA_IM3D **copie );
void empty_data_im3d ( DATA_IM3D *im3d );
void empty_data_pt3d ( DATA_PT3D *pt3d );
int extrait_sous_im3d ( DATA_IM3D* im3d, DATA_IM3D* sous_im3d, int i_min, int i_max, int j_min, int j_max );
int filtre_ima1_im3d ( DATA_IM3D* im3d, unsigned char* q_image, int nblig, int nbcol );
int filtre_im3d_ima1 ( DATA_IM3D* im3d, unsigned char* q_image, int nblig, int nbcol );
int im3d_2_plan ( DATA_IM3D* im3d, double *abcE, int *type );
int info_data_im3d ( DATA_IM3D* im3d );
int info_data_pt3d ( DATA_PT3D* pt3d );
int info_im3d_pt3d ( DATA_IM3D* im3d, int i, int j );

int init_data_im3d ( DATA_IM3D *im3d, int nblig, int nbcol, int covar, int normales, int size_local );
int init_data_pt3d ( DATA_PT3D *new_pt3d, int covar, int normales, int size_local );
int read_data_im3d_ascii ( DATA_IM3D* im3d, char* nom_fichier );
int read_data_im3d_binaire ( DATA_IM3D* im3d, char* nom_fichier );

int read_im3d_header ( FILE* fichier, DATA_IM3D_HEADER* im3d_header );

int seuille_im3d ( DATA_IM3D* im3d, int flag_1, double seuil_1, int flag_2, double seuil_2, int flag_3, double seuil_3 );
int sous_echantillonne_im3d ( DATA_IM3D* im3d, int taux_reduction, DATA_IM3D* im3d_reduite );
int write_data_im3d_ascii ( DATA_IM3D* im3d, char* nom_fichier);
int write_data_im3d_binaire ( DATA_IM3D* im3d, char* nom_fichier);
int write_data_im3d_off ( DATA_IM3D* im3d, char* nom_fichier, double ATHRESH, double RTHRESH, unsigned char* q_im_gauche );

int *	Delaunay_2d(int *NbTri, double *X, double *Y, int SizeX_Y);

#ifdef __cplusplus
}
#endif
#endif /* CALIFE_LIBIMAGES3D_H */
